/////////////////////////////////////////////////////////////////////////////
/*
  Copyright 2001 Ronald S. Burkey.
  Latex support Copyright 2001 Joe Cherry.

  This file is part of GutenMark.

  GutenMark is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GutenMark is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GutenMark; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	AutoMark.c
  Purpose:	This is the main program for AutoMark, a vanilla-ANSI C 
  		program that marks up a vanilla-ASCII Project Gutenberg 
		(or similar) text file to HTML, with the intention of 
		making it printable on demand, or very beautifully readable.  
		In other words, Project Gutenberg has provided the content, 
		but now I want to make that content aesthetically pleasing.
  Mods:		08/30/01 RSB	Began.
  		09/03/01 RSB	Added MarkInsertChar.
		09/04/01 RSB	Some things that produced invalid HTML were 
				fixed:  '>', '<', '&' within body of the 
				text.  Empty <head></head> (no <title>).
				Empty anchor after Gutenberg header.  Open 
				paragraphs, headings, block quotes at end of 
				file.	
		09/06/01 RSB	Now allow output file as command-line 
				argument.  Also, allow some command-line 
				options (--no-justify, --help).  Added a lot 
				more processing on the "title".		
		09/08/01 RSB	Area between the PG header and the actual text, 
				which I call the "prefatory" area, is now now 
				detected and used essentially without markup.  
				Phrases found in this area (often a table of 
				contents) are used preferentially to detect 
				headings in the text and (conversely) to 
				reject identifications as headings.  MANY 
				additional tweaks performed.  Tables are now 
				handled, but only as "preformatted text" and 
				not a HTML tables.
		09/09/01 RSB	Again, numerous changes.
		09/15/01 RSB	Now handles PG so-called "8-bit ASCII".  A 
				lot more cleanup, such as block quotes, new 
				types of italic styles, and so on.
		09/16/01 RSB	Here is just a trivial remark added for 
				checking CVS.
		11/02/01 RSB	Added GPL disclaimer, and cleaned up the 
				formatting, prior to first web release.	
				Added startup messages.	
		11/09/01 RSB	Convert strings of dashes (>1) to mdashes
				and ndashes.	
		11/12/01 RSB	Added the --yes-header option,
				and PG header/ender stuff.
		11/13/01 RSB	Added smart quotes.
		11/19/01 RSB	Added first cut at ALL-CAPS italicizing.
				Added a GutenMark declaration at the 
				beginning of the output.
		11/24/01 RSB	Added --no-foreign.
		11/25/01 RSB	Fixed bug in parsing the --profile switch.
				Added --no-diacritical switch.
				Began adding Joe Cherry's --latex patch.
				(Still need to modify the patch to handle
				the GutenMark disclaimer.)
		11/26/01 RSB	Completed Latex patch.  Joe indicates he's
				working on some of the problems I've pointed
				out.	
		11/29/01 RSB	Added wordlist output (GutenMark.native &
				GutenMark.foreign) to --debug output.
								
  Problems:	>	Greek phrases transliterated (without markup) into 
  			English.
		>	Footnotes embedded directly in the text, in braces 
			but without other markup, thusly:  
			"[text of footnote]"
		>	Footnotes in a "curly brace" style, collected at the 
			end of the text, thusly:  in the text body "{10}", 
			and at the end "{10} text of note".
		>	Should stuff like "to-"/"day" be  replaced 
			by "today" or by "to-day"?  The former requires
			building a spelling dictionary into the program.
			But in many circa-1900 documents, "to-day"
			and "to-morrow" are the actual word-forms
			expected (not just at the ends of lines).
		>	The preparer of the etext using in error
			a single dash without spaces to indicate a break
			in thought.  In other words, "-" where "--" or
			" -- " would be more appropriate.  There is no
					way of detecting this problem.
		>	There is no way to correctly recognize things like
			"right- and left-hand" (when "right-" appears at 
			the end of a line).  In other words, it would be
			converted to "right-and left-hand".  (This doesn't
			actually occur in the text, by the way.)
		>	Footnotes with markings in the body of the
			text like "[FN#10]", and with the footnotes 
			themselves collected at the end, like
			"[FN#10] text of footnote".
		>	Advanced treatment of style in which italicized text 
			is in all-caps.  Here are some PROPOSED improvements:
			1.	All-cap Roman Numbers will not be converted to 
				italics unless italicized text appears on one 
				or the other side of it as well.  "I" won't be 
				italicized unless surrounded by other 
				italicized text.  In either case, they will not 
				be converted to lower case.
			2.	Database of book titles.					
		>	Footnotes marked with "*", "**", etc., and collected 
			at the ends of chapters.
		>	Footnotes embedded in the text, but with numbers:  
			"[53]" in the text, and a line like "[53] Text of 
			footnote." after the end of the paragraph.
		>	Tabular data:  as in flymc10.txt.
		>	Endnote style marked in the text as (a), (b), (c), 
			etc.
		>	Alignment, as sometimes found in plays (or even in 
			verse, like 8faun10.txt) where the left half and 
			right half of the lines are separated by braces, 
			represented by {, |, (, or [ where multiple lines 
			are expected to line up.  For example:
						   ( Nonsense!
						   | You can't be serious!
			I've decided to leave you! { What!
						   | You haven't the nerve!
						   ( That's crazy talk!	
			In other words, the (, |, and { symbols are expected 
			to form a single large brace.
		>	Could use a better way to detect indented verse.
			Really should analyze the indentation patterns of 
			the stanzas, so that when something weird (but 
			indented properly) like a leading lower-case character 
			is encountered, it's treated properly.
  
  The program has to handle many special cases, such as differing paragraph 
  styles (indentation vs. no indentation, extra blank lines between paragraphs
  vs. no extra blank lines, etc.), differing styles for indicating italics 
  (or bold, or underlining), hyphenation vs. no hyphenation, Gutenberg vs. 
  non-Gutenberg, etc.  However, I want this program to be AUTOMATIC, so I'm 
  holding off on adding command-line switches for ANYTHING.  The program will 
  analyze the text to determine the styles, and if extra passes are necessary 
  to accomplish this, then so be it!
  
  The program works as follows:
  	AutoMark [options] [inputfile [outputfile]]
  If outputfile and/or inputfile are missing, then the stdout and/or stdin 
  are used.	
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <zlib.h>
#include "AutoMark.h"

void NormalizeTitle (char *s, int size);


//--------------------------------------------------------------------------

int
main (int argc, char *argv[])
{
#define MAX_FREQUENT 100
  struct
  {
    int Index;
    int Count;
  }
  Frequents[MAX_FREQUENT];
  AnalysisDataset Dataset = { NULL };
  int ErrorCode, c, lastc, Remove, AtEnd, i, j, k;
  MarkupRecord Mark;
  unsigned long Offset;
  char s[256], Profile[64] = "english";
  int FilesFound = 0;
  int inchapter = 0;
  int sentpgheader = 0;
  FILE *OutputFile;
  char *InputFilename = NULL;
  int Removed = -1;

  // Process the command line.
  Dataset.InputFile = stdin;
  OutputFile = stdout;
  fprintf (stderr, "GutenMark PG-to-HTML markup tool, build " __DATE__ "\n");
  fprintf (stderr,
	   "Copyright (C) 2001 Ronald Burkey (info@birdsproject.org)\n");
  fprintf (stderr,
	   "Latex output Copyright (C) 2001 Joe Cherry (joe@vimaster.org)\n");
  fprintf (stderr, "GutenMark comes with ABSOLUTELY NO WARRANTY.\n");
  fprintf (stderr, "This is free software, licensed under the GNU GPL.\n");
  for (i = 1; i < argc; i++)
    {
      if (*argv[i] == '-')
	{
	  if (0 == strcmp (argv[i], "--help"))
	    {
	      printf
		("----------------------------------------------------\n");
	      printf ("USAGE:");
	      printf
		("\tAutoMark [Options] [InputTxtFile [OutputHtmlFile]]\n");
	      printf
		("If the output file and/or input file are missing, "
		 " then the standard\n");
	      printf ("output and/or input are used.  "
		      "The available options are:\n");
	      printf ("\t--debug          Create optional logfiles.\n");
	      printf ("\t--help           Displays this information.\n");
	      printf ("\t--latex          Output LaTex instead of html.\n");
	      printf
		("\t--no-diacritical Turn off restoration of diacritical\n"
		 "\t                 marks.  (On by default.)\n");
	      printf
		("\t--no-foreign     Turn off automatic italicizing of\n"
		 "\t                 foreign text.  (On by default.)\n");
	      printf
		("\t--no-justify     Allows ragged right margin.\n"
		 "\t                 Default is justified.\n");
	      printf
		("\t--no-mdash       Keeps strings of dashes intact.\n"
		 "\t                 Default converts to mdashes/ndashes.\n");
	      printf
		("\t--profile=xxx    Select language xxx.  Actually, these\n"
		 "\t                 names refer not to languages, but rather\n"
		 "\t                 to names of profiles in GutenMark.cfg.\n"
		 "\t                 But the profiles are usually named for\n"
		 "\t                 a particular language.\n");
	      printf
		("\t--yes-header     Causes PG file-header to be retained.\n");
	      return (0);
	    }
	  else if (0 == strcmp (argv[i], "--no-justify"))
	    {
	      Dataset.NoJustify = 1;
	      continue;
	    }
	  else if (0 == strcmp (argv[i], "--no-mdash"))
	    {
	      Dataset.NoMdash = 1;
	      continue;
	    }
	  else if (0 == strcmp (argv[i], "--yes-header"))
	    {
	      Dataset.YesHeader = 1;
	      continue;
	    }
	  else if (0 == strcmp (argv[i], "--no-foreign"))
	    {
	      Dataset.NoForeign = 1;
	      continue;
	    }
	  else if (0 == strcmp (argv[i], "--no-diacritical"))
	    {
	      Dataset.NoDiacritical = 1;
	      continue;
	    }
	  else if (0 == strcmp (argv[i], "--debug"))
	    {
	      Dataset.LogFile = fopen ("GutenMark.log", "w");
	      continue;
	    }
	  else if (0 == strncmp (argv[i], "--profile=", 10))
	    {
	      if (strlen (&argv[i][10]) < sizeof (Profile))
		strcpy (Profile, &argv[i][10]);
	      continue;
	    }
	  else if (0 == strcmp (argv[i], "--latex"))
	    {
	      Dataset.Latex = 1;
	      continue;
	    }
	  else
	    {
	      fprintf (stderr, "Unknown command-line parameter: %s\n"
		       "Use \"GutenMark --help\" for more info.\n", argv[i]);
	      continue;
	    }
	}
      else if (FilesFound == 0)
	{
	  Dataset.InputFile = fopen (argv[i], "rb");
	  if (Dataset.InputFile == NULL)
	    {
	      fprintf (stderr, "Input file does not exist.\n");
	      return (1);
	    }
	  InputFilename = argv[i];
	  FilesFound++;
	}
      else if (FilesFound == 1)
	{
	  OutputFile = fopen (argv[i], "w+b");
	  if (OutputFile == NULL)
	    {
	      fprintf (stderr, "Output file cannot be created.\n");
	      return (1);
	    }
	  FilesFound++;
	}
      else
	{
	  fprintf (stderr, "Too many filename listed on command line.\n");
	  fclose (Dataset.InputFile);
	  fclose (OutputFile);
	  return (1);
	}
    }

  // Create in-memory wordlist from the etext.
  Dataset.Words = CreateWordlist (Dataset.InputFile);
  // Determine the most frequently used words in the etext, 
  // from the wordlist just created.
  for (i = 0; i < MAX_FREQUENT; i++)
    Frequents[i].Count = -1;
  k = Frequents[0].Count;	// Min Count currently in Frequents[].  
  for (j = 0; j < Dataset.Words->NumWords; j++)
    if (Dataset.Words->Words[j].Count > k)
      {
	// The word has been used frequently enough to displace one
	// of the words already in Frequents[].  The words in 
	// Frequents[] are in the order of least frequent (at 0)
	// to most frequent (at MAX_FREQUENT-1).
	for (i = 0; i < MAX_FREQUENT - 1 &&
	     Frequents[i + 1].Count < Dataset.Words->Words[j].Count; i++)
	  Frequents[i] = Frequents[i + 1];
	Frequents[i].Count = Dataset.Words->Words[j].Count;
	Frequents[i].Index = j;
	k = Frequents[1].Count;
      }
  for (j = MAX_FREQUENT - 1; j >= 0; j--)
    if (Frequents[j].Count != -1)
      {
	Dataset.Words->Words[Frequents[j].Index].Frequent = 1;
	if (Dataset.LogFile != NULL)
	  fprintf (Dataset.LogFile, "Index=%d Count=%d Word=\"%s\"\n",
		   Frequents[j].Index, Frequents[j].Count,
		   Dataset.Words->Words[Frequents[j].Index].Full);
      }
  // Read the global wordlists and namelists.
  if (0 != MatchWordlists (Dataset.LogFile, Dataset.Words, Profile))
    {
      fprintf (stderr, "Error matching wordlists/namelists.\n");
      return (1);
    }
  if (Dataset.LogFile != NULL)
    {
      gzFile Native, Foreign;
      Native = gzopen ("GutenMark.native.gz", "w");
      Foreign = gzopen ("GutenMark.foreign.gz", "w");
      fprintf (Dataset.LogFile, "NumWords = %d\n", Dataset.Words->NumWords);
      for (i = 0; i < Dataset.Words->NumWords; i++)
	{
	  if (0 != (Dataset.Words->Words[i].WordlistStatus & SPELL_NATIVE)
	      && Native != NULL)
	    {
	      if (Dataset.Words->Words[i].Match != NULL)
		gzprintf (Native, "%s\n", Dataset.Words->Words[i].Match);
	      else
		gzprintf (Native, "%s\n", Dataset.Words->Words[i].Full);
	    }
	  if (0 != (Dataset.Words->Words[i].WordlistStatus & SPELL_NONNATIVE)
	      && Foreign != NULL)
	    {
	      if (Dataset.Words->Words[i].Match != NULL)
		gzprintf (Foreign, "%s\n", Dataset.Words->Words[i].Match);
	      else
		gzprintf (Foreign, "%s\n", Dataset.Words->Words[i].Full);
	    }
	  fprintf (Dataset.LogFile, "Normalized=");
	  if (Dataset.Words->Words[i].Normalized == NULL)
	    fprintf (Dataset.LogFile, "NULL");
	  else
	    fprintf (Dataset.LogFile, "%s",
		     Dataset.Words->Words[i].Normalized);
	  fprintf (Dataset.LogFile, ", Full=");
	  if (Dataset.Words->Words[i].Full == NULL)
	    fprintf (Dataset.LogFile, "NULL");
	  else
	    fprintf (Dataset.LogFile, "%s", Dataset.Words->Words[i].Full);
	  if (Dataset.Words->Words[i].Match == NULL)
	    /* fprintf (Dataset.LogFile, ", Match=NULL") */ ;
	  else
	    fprintf (Dataset.LogFile, ", Match=%s",
		     Dataset.Words->Words[i].Match);
	  fprintf (Dataset.LogFile, ", NotAtBeginning=%d, Status=%02X\n",
		   Dataset.Words->Words[i].NotAtBeginning,
		   Dataset.Words->Words[i].WordlistStatus);
	}
      if (Native != NULL)
	gzclose (Native);
      if (Foreign != NULL)
	gzclose (Foreign);
    }
  fseek (Dataset.InputFile, 0, SEEK_END);
  Dataset.FileEnderRegion = ftell (Dataset.InputFile) - 500;
  fseek (Dataset.InputFile, 0, SEEK_SET);
  Dataset.MarkupFile = tmpfile ();
  if (Dataset.MarkupFile == NULL)
    {
      fprintf (stderr, "Cannot create temporary markup file.\n");
      return (1);
    }

  // First stage:  Style analysis.  The analysis actually performs all 
  // of the markup, but doesn't modify the input or output files.  
  // Instead the  markups are saved to another file.
  ErrorCode = Markup (&Dataset);
  if (ErrorCode != 0)
    {
      fprintf (stderr, "An error code of %d occurred during markup.\n",
	       ErrorCode);
      return (1);
    }

  // Second stage:  Print output header.
  if (Dataset.Latex == 1)
    fprintf (OutputFile, "\\documentclass{book}\n");
  else
    {
      fprintf
	(OutputFile,
	 "<!doctype html public \"-//w3c//dtd html 4.0 transitional//en\">\n");
      fprintf (OutputFile, "<html>\n");
      fprintf (OutputFile, "<head>\n");
    }
  fseek (Dataset.InputFile, 0, SEEK_SET);
  fgets (s, sizeof (s) - 1, Dataset.InputFile);
  fgets (s + strlen (s), sizeof (s) - strlen (s) - 1, Dataset.InputFile);
  NormalizeTitle (s, sizeof (s));
  if (Dataset.Latex == 1)
    {
      fprintf (OutputFile, "\\title{%s}\n", s);
      fprintf (OutputFile, "\\begin{document}\n");
      fprintf (OutputFile, "\\maketitle\n");
      fprintf (OutputFile, "\\tableofcontents\n");
    }
  else
    {
      fprintf (OutputFile, "<TITLE>%s</TITLE>\n", s);
      fprintf (OutputFile, "</head>\n");
      fprintf (OutputFile, "<body>\n");
      fprintf (OutputFile, "<!-- Short-line cutoffs are %u and %u -->\n",
	       Dataset.ShortLineSize, Dataset.ReallyShortLineSize);
    }

  // Third stage:  Print marked-up text.  This is basically a matter 
  // of merging the markup file with the input file.
  fseek (Dataset.MarkupFile, 0, SEEK_SET);
  fseek (Dataset.InputFile, 0, SEEK_SET);
  if (Dataset.YesHeader)
    {
      ErrorCode = fread (&Mark, sizeof (Mark), 1, Dataset.MarkupFile);
      if (ErrorCode != 1)
	Mark.Type = MarkNoMoreMarks;
      Offset = 0;
    }
  else
    {
      do
	{
	  ErrorCode = fread (&Mark, sizeof (Mark), 1, Dataset.MarkupFile);
	  if (ErrorCode != 1)
	    {
	      Mark.Type = MarkNoMoreMarks;
	      break;
	    }
	}
      while (Mark.Offset < Dataset.TextStart);
      fseek (Dataset.InputFile, Dataset.TextStart, SEEK_SET);
      Offset = Dataset.TextStart;
    }
  for (Remove = AtEnd = c = 0;
       (lastc = c, c = fgetc (Dataset.InputFile)) != EOF; Offset++)
    {
    LastShot:
      while (Offset == Mark.Offset && Mark.Type != MarkNoMoreMarks)
	{
	  switch (Mark.Type)
	    {
	    case MarkTolower:
	      //if (Removed != Offset)
	      putc (DiacriticalTolower (c), OutputFile);
	      Remove = 1;
	      break;
	    case MarkToupper:
	      //if (Removed != Offset)
	      putc (DiacriticalToupper (c), OutputFile);
	      Remove = 1;
	      break;
	    case MarkBeginSmartQuote:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "``");
	      else
		fprintf (OutputFile, "&#8220;");
	      Remove = 1;
	      break;
	    case MarkEndSmartQuote:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "''");
	      else
		fprintf (OutputFile, "&#8221;");
	      Remove = 1;
	      break;
	    case MarkLsquo:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "`");
	      else
		fprintf (OutputFile, "&#8216;");
	      Remove = 1;
	      break;
	    case MarkRsquo:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "'");
	      else
		fprintf (OutputFile, "&#8217;");
	      Remove = 1;
	      break;
	    case MarkRemoveChar:
	      Remove = 1;
	      Removed = Offset;
	      break;
	    case MarkInsertChar:
	      putc (Mark.Insert, OutputFile);
	      break;
	    case MarkNbsp:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "~");
	      else
		fprintf (OutputFile, "&nbsp;");
	      break;
	    case MarkBeginJustifiedParagraph:
	      if (Dataset.NoJustify)
		{
		  if (Dataset.Latex == 1)
		    fprintf (OutputFile, "\n\n");
		  else
		    fprintf (OutputFile, "<P>");
		}
	      else
		{
		  if (Dataset.Latex == 1)
		    fprintf (OutputFile, "\n\n");
		  else
		    fprintf (OutputFile, "<P align=justify>");
		}
	      break;
	    case MarkBeginRaggedParagraph:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\n\n");
	      else
		fprintf (OutputFile, "<P>");
	      break;
	    case MarkBeginCenteredParagraph:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\n\n");
	      else
		fprintf (OutputFile, "<P align=center>");
	      break;
	    case MarkEndParagraph:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\n\n");
	      else
		fprintf (OutputFile, "</P>");
	      break;
	    case MarkBeginItalics:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\\textit{");
	      else
		fprintf (OutputFile, "<I>");
	      break;
	    case MarkEndItalics:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "}");
	      else
		fprintf (OutputFile, "</I>");
	      break;
	    case MarkBeginBold:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\\textbf{");
	      else
		fprintf (OutputFile, "<B>");
	      break;
	    case MarkEndBold:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "}");
	      else
		fprintf (OutputFile, "</B>");
	      break;
	    case MarkBeginUnderline:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\\emph{");
	      else
		fprintf (OutputFile, "<U>");
	      break;
	    case MarkEndUnderline:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "}");
	      else
		fprintf (OutputFile, "</U>");
	      break;
	    case MarkBreak:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\n\n");
	      else
		fprintf (OutputFile, "<BR>");
	      break;
	    case MarkBlockquote:	// I think, not used.
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\\begin{quotation}\n");
	      else
		fprintf (OutputFile, "<BLOCKQUOTE>");
	      break;
	    case MarkEndBlockquote:	// I think, not used.
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\\end{quotation}\n");
	      else
		fprintf (OutputFile, "</BLOCKQUOTE>");
	      break;
	    case MarkHeader1:
	      if (Dataset.Latex == 1)
		{
		  fprintf (OutputFile, "\\chapter{");
		  inchapter = 1;
		}
	      else
		fprintf (OutputFile, "<H1>");
	      break;
	    case MarkEndHeader1:
	      if (Dataset.Latex == 1)
		{
		  fprintf (OutputFile, "}");
		  inchapter = 0;
		}
	      else
		fprintf (OutputFile, "</H1>");
	      break;
	    case MarkBeginSubtitle:
	      if (Dataset.Latex == 1)
		{
		  fprintf (OutputFile, "\\section{");
		  inchapter = 1;
		}
	      else
		fprintf (OutputFile, "<P><B>");
	      break;
	    case MarkEndSubtitle:
	      if (Dataset.Latex == 1)
		{
		  fprintf (OutputFile, "}");
		  inchapter = 0;
		}
	      else
		fprintf (OutputFile, "</B></P>");
	      break;
	    case MarkBeginTable:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\\begin{verbatim}\n");
	      else
		fprintf (OutputFile, "<PRE>");
	      break;
	    case MarkEndTable:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\\end{verbatim}\n");
	      else
		fprintf (OutputFile, "</PRE>");
	      break;
	    case MarkInsertMdash:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "--");
	      else
		fprintf (OutputFile, "&#8212;");
	      break;
	    case MarkInsertNdash:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "-");
	      else
		fprintf (OutputFile, "&#8211;");
	      break;
	    case MarkSoftHyphen:
	      if (Dataset.Latex == 1)
		fprintf (OutputFile, "\\-");
	      else
		fprintf (OutputFile, "&#173;");
	      break;
	    case MarkJumpPastGutenbergHeader:
	      if (Dataset.Latex == 1)
		{
		  fprintf (OutputFile,
			   "\\chapter{Project Gutenberg Fine Print}\\begin{verbatim}\n");
		  sentpgheader = 1;
		}
	      else
		{
		  fprintf (OutputFile,
			   "<H1>Project Gutenberg Fine Print</H1><pre>\n");
		}
	      break;
	    case MarkEndOfGutenbergHeader:
	      if (Dataset.Latex == 1)
		{
		  if (sentpgheader == 1)
		    fprintf (OutputFile, "\\end{verbatim}\n");
		  sentpgheader = 0;
		  if (Dataset.LowestNonPrefatoryLine > 0)
		    {
		      fprintf (OutputFile,
			       "\\chapter*{Prefatory Materials}\n"
			       "\\begin{quotation}\n"
			       "\\textit{This is a modified "
			       "etext created by \\textbf{GutenMark} "
			       "software, version " __DATE__ ".");
		      if (InputFilename != NULL)
			fprintf (OutputFile,
				 "~ The original text was contained "
				 "in a computer file named \"%s.\"",
				 InputFilename);
		      fprintf (OutputFile, "~ Any comments below "
			       "about the preparation of the text "
			       "refer to the \\emph{original} "
			       "etext file, and do not directly "
			       "refer to this modified version of "
			       "the text.~ In particular, no individuals "
			       "named below bear any responsibility for "
			       "changes to the text.}\n"
			       "\\end{quotation}\n");
		    }
		}
	      else
		{
		  fprintf (OutputFile, "</pre>");
		  if (Dataset.LowestNonPrefatoryLine > 0)
		    fprintf (OutputFile, "<H1>Prefatory Materials</H1>\n");
		  fprintf (OutputFile,
			   "<BLOCKQUOTE><I><FONT SIZE=-1>This is a modified "
			   "etext created by <b>GutenMark</b> software, "
			   "version " __DATE__ ".");
		  if (InputFilename != NULL)
		    fprintf (OutputFile,
			     "&nbsp; The original text was contained in a "
			     "computer file named \"%s.\"", InputFilename);
		  fprintf (OutputFile, "&nbsp; Any comments below about the "
			   "preparation of the text refer to the "
			   "<U>original</U> etext file, and do not directly "
			   "refer to this modified version of the text.&nbsp; "
			   "In particular, no individuals named below bear "
			   "any responsibility for changes to the text."
			   "</FONT></I></BLOCKQUOTE>\n");
		}
	      break;
	    case MarkGutenbergEnder:
	      if (Dataset.YesHeader)
		{
		}
	      else
		AtEnd = 1;
	      break;
	    case MarkNoMoreMarks:
	      break;
	    }
	  ErrorCode = fread (&Mark, sizeof (Mark), 1, Dataset.MarkupFile);
	  if (ErrorCode != 1)
	    Mark.Type = MarkNoMoreMarks;
	}
      if (AtEnd)
	break;
      if (Remove)
	Remove = 0;
      else
	{

	  if (Dataset.Latex == 1)
	    {
	      if (c == '$')
		fprintf (OutputFile, "\\$");
	      else if (c == '&')
		fprintf (OutputFile, "\\&");
	      else if (c == '%')
		fprintf (OutputFile, "\\%%");
	      else if (c == '#')
		fprintf (OutputFile, "\\#");
	      else if (c == '_')
		fprintf (OutputFile, "\\_");
	      else if (c == '{')
		fprintf (OutputFile, "\\{");
	      else if (c == '}')
		fprintf (OutputFile, "\\}");
	      else if (c == '~')
		fprintf (OutputFile, "\\~");
	      else if (c == '^')
		fprintf (OutputFile, "\\^");
	      else if (c == '\\')
		fprintf (OutputFile, "$\\backslash$");
	      else
		{
		  if (inchapter == 1 && (c == '\r' || c == '\n'))
		    putc (' ', OutputFile);
		  else
		    putc (c, OutputFile);
		}
	    }
	  else
	    {
	      if (c == '<')
		fprintf (OutputFile, "&lt;");
	      else if (c == '>')
		fprintf (OutputFile, "&gt;");
	      else if (c == '&')
		fprintf (OutputFile, "&amp;");
	      //else if (isspace (c))
	      //  putc (' ', OutputFile);
	      //else if (c < ' ')
	      //  ;    
	      else if (c < 128)
		putc (c, OutputFile);
	      else
		fprintf (OutputFile, "&#%d;", c);
	    }
	}
    }
  // What this does is to allow us to process any remaining markups that
  // are supposed to occur at the very end of the input (such as </p>).  
  if (AtEnd == 0)
    {
      AtEnd = 1;
      goto LastShot;
    }

  // Last stage:  Close the output HTML stream.          
  if (Dataset.Latex == 1)
    {
      fprintf (OutputFile, "\\end{document}\n");
    }
  else
    {
      fprintf (OutputFile, "</body>\n");
      fprintf (OutputFile, "</html>\n");
    }
  fclose (Dataset.MarkupFile);
  fclose (Dataset.InputFile);
  fclose (OutputFile);
  return (0);

}

//---------------------------------------------------------------------
// The "title" has been read from the first line of the file.  
// However, it may have all kinds of Gutenbergy stuff, like asterisks, 
// the phrase "Project Gutenberg etext of", and so on.

void
NormalizeTitle (char *s, int size)
{
  char *ss;

  // Make sure the string is null-terminated.
  s[size - 1] = '\0';

  // Get rid of leading asterisks and white space.
  for (ss = s; *ss == '*' || isspace (*ss); ss++);
  if (ss > s)
    strcpy (s, ss);

  // Let's get rid of trailing asterisks and white space also.
  if (*s != '\0')
    {
      for (ss = s + strlen (s) - 1; *ss == '*' || isspace (*ss); ss--);
      ss[1] = '\0';
    }

  // Let's also get rid of multiple spaces.
  for (ss = s; *ss;)
    {
      if (isspace (*ss) && isspace (ss[1]))
	strcpy (ss, ss + 1);
      else if (*ss == '\t')
	*ss = ' ';
      else
	ss++;
    }

  // Look for some stuff that PG sometimes adds AFTER the title.
  for (ss = s; *ss; ss++)
    if (*ss == '*' || *ss == '#' || *ss == '(' || *ss == '[' || *ss == '{')
      {
	while (ss > s
	       && (ss[-1] == ' ' || ss[-1] == ',' || ss[-1] == '.'
		   || ss[-1] == ';' || ss[-1] == ':' || ss[-1] == '-'))
	  ss--;
	*ss = '\0';
	break;
      }

  // Now get rid of several common things that PG likes to add:
  if (0 == strncasecmp (s, "PG ", 3))
    strcpy (s, s + 3);
  if (0 == strncasecmp (s, "PG\'s ", 5))
    strcpy (s, s + 5);
  else if (0 == strncasecmp (s, "Project Gutenberg\'s ", 20))
    strcpy (s, s + 20);
  else if (0 == strncasecmp (s, "Project Gutenberg ", 18))
    strcpy (s, s + 18);
  else if (0 == strncasecmp (s, "The Project Gutenberg ", 22))
    strcpy (s, s + 22);
  if (0 == strncasecmp (s, "etext of ", 9))
    strcpy (s, s + 9);
  else if (0 == strncasecmp (s, "etext: ", 7))
    strcpy (s, s + 7);
  else if (0 == strncasecmp (s, "etext, ", 7))
    strcpy (s, s + 7);
  else if (0 == strncasecmp (s, "etext; ", 7))
    strcpy (s, s + 7);
  else if (0 == strncasecmp (s, "etext. ", 7))
    strcpy (s, s + 7);
  else if (0 == strncasecmp (s, "etext ", 6))
    strcpy (s, s + 6);

  // Finally, if the string is empty, fill in a default value.
  if (*s == '\0')
    strcpy (s, "A Project Gutenberg Book");

}
